#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdbool.h>
#include <stdint.h>
#include <getopt.h>

#include <ev.h>
#include <citrusleaf_event/ev2citrusleaf.h>

// ----------------------------------------------------------------------------
// Local logging macros
// ----------------------------------------------------------------------------
#define STR(x) #x
#define XSTR(x) STR(x)
#define LOG(...) {                                                        \
  printf("[" __FILE__ ":" XSTR(__LINE__) "] " __VA_ARGS__); printf("\n"); \
  fflush(stdout);                                                         \
}

// ----------------------------------------------------------------------------
// Constants
// ----------------------------------------------------------------------------
const char DEFAULT_HOST[] = "127.0.0.1";
const int DEFAULT_PORT = 3000;
const char DEFAULT_NAMESPACE[] = "test";
const char DEFAULT_SET[] = "test-set";
const int DEFAULT_TIMEOUT_MSEC = 200;

const int CLUSTER_VERIFY_TRIES = 5;
const __useconds_t CLUSTER_VERIFY_INTERVAL = 1000 * 1000; // 1 second

const char KEY_STRING[] = "test-key";
const uint8_t BLOB[] = {
    0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF
};


// ----------------------------------------------------------------------------
// Event Plugin
// ----------------------------------------------------------------------------
typedef void (*handler_t)(int, short, void*);

typedef struct _ev_plugin_cookie {
  struct ev_loop* loop;
  bool need_cleanup;
} ev_plugin_cookie_t;

typedef struct _ev_callback_cookie {
  handler_t cl_lib_callback;
  void* cl_lib_data;
} ev_callback_cookie_t;

ev2citrusleaf_ev_operations* create_event_plugin(struct ev_loop*);
void destroy_event_plugin(ev2citrusleaf_ev_operations*);

void* create_timer(ev2citrusleaf_ev_operations*);
void delete_timer(ev2citrusleaf_ev_operations*, void*);
void destroy_timer(ev2citrusleaf_ev_operations*, void*);
int update_timer(ev2citrusleaf_ev_operations*, void*, int, void*, handler_t);
void timer_callback(struct ev_loop*, ev_timer*, int);

void* create_event(ev2citrusleaf_ev_operations*);
void delete_event(ev2citrusleaf_ev_operations*, int, void*);
void destroy_event(ev2citrusleaf_ev_operations*, void*);
int update_event(ev2citrusleaf_ev_operations*, int, void*, short, void*,
    handler_t);
void event_callback(struct ev_loop*, ev_io*, int);

void run_event_loop(ev2citrusleaf_ev_operations*);
void stop_event_loop(ev2citrusleaf_ev_operations*);

short libev_events_to_cl(int);
int cl_events_to_libev(short);

// ----------------------------------------------------------------------------
// Config and command line options
// ----------------------------------------------------------------------------
typedef struct _config {
  const char* host;
  int port;
  const char* namespace;
  const char* set;
  int timeout_msec;
} config_t;

config_t g_config;

static bool set_config(int, char**);
static void destroy_config();
static void usage() {}

// ----------------------------------------------------------------------------
// CitrusLeaf
// ----------------------------------------------------------------------------
ev2citrusleaf_cluster* g_cluster = NULL;
ev2citrusleaf_object g_key;

typedef struct _verify_cluster_callback_cookie {
  ev2citrusleaf_ev_operations* event_ops;
  void* timer;
} verify_cluster_callback_cookie_t;

static bool start_cluster_management(ev2citrusleaf_ev_operations*);
static void stop_cluster_management();
static void verify_cluster_callback(int, short, void*);


// ============================================================================
// Main
// ============================================================================
int main(int argc, char** argv) {
  if (!set_config(argc, argv)) {
    destroy_config();
    return -1;
  }

  cf_set_log_level(CF_WARN);

  ev2citrusleaf_ev_operations* ev = create_event_plugin(NULL);

  LOG("INFO: Starting cluster management");
  if (!start_cluster_management(ev)) {
    return -1;
  }

  LOG("INFO: Starting event loop");
  ev->run_event_loop(ev);

  stop_cluster_management();
  free(ev);

  return 0;
}

// ============================================================================
// Event Plugin Implementation
// ============================================================================
ev2citrusleaf_ev_operations* create_event_plugin(struct ev_loop* loop) {
  ev2citrusleaf_ev_operations* ev_ops = (ev2citrusleaf_ev_operations*)
      malloc(sizeof (ev2citrusleaf_ev_operations));
  if (!ev_ops) {
    LOG("ERROR: unable to allocate memory for event plugin");
    return NULL;
  }

  ev_plugin_cookie_t* cookie = (ev_plugin_cookie_t*)
      malloc(sizeof (ev_plugin_cookie_t));
  if (!cookie) {
    LOG("ERROR: unable to allocate memory for event plugin cookie");
    free(ev_ops);
    return NULL;
  }

  memset(ev_ops, 0, sizeof (ev2citrusleaf_ev_operations));
  memset(cookie, 0, sizeof (ev_plugin_cookie_t));

  cookie->need_cleanup = false;
  if (!loop) {
    loop = ev_loop_new(EVBACKEND_EPOLL | EVFLAG_NOENV |
        EVFLAG_NOSIGMASK);
    cookie->need_cleanup = true;
  }
  cookie->loop = loop;
  ev_ops->cookie = (void*) cookie;

#ifdef DEBUG
  ev_verify(loop);
#endif /* DEBUG */

  ev_ops->need_cleanup = 1;
  ev_ops->destructor = destroy_event_plugin;

  ev_ops->create_timer = create_timer;
  ev_ops->delete_timer = delete_timer;
  ev_ops->destroy_timer = destroy_timer;
  ev_ops->update_timer = update_timer;

  ev_ops->create_event = create_event;
  ev_ops->delete_event = delete_event;
  ev_ops->destroy_event = destroy_event;
  ev_ops->update_event = update_event;

  ev_ops->run_event_loop = run_event_loop;
  ev_ops->stop_event_loop = stop_event_loop;

  return ev_ops;
}

void destroy_event_plugin(ev2citrusleaf_ev_operations* ev_ops) {
  if (!ev_ops)
    return;

  ev_plugin_cookie_t* cookie = (ev_plugin_cookie_t*) ev_ops->cookie;
  if (cookie) {
    if (cookie->need_cleanup) {
      ev_loop_destroy(cookie->loop);
    }
#ifdef DEBUG
    memset(cookie, 0, sizeof (ev_plugin_cookie_t));
#endif /* DEBUG */
    free(cookie);
  }
#ifdef DEBUG
  memset(ev_ops, 0, sizeof (ev2citrusleaf_ev_operations));
#endif /* DEBUG */
  free(ev_ops);
}

void* create_timer(ev2citrusleaf_ev_operations* ev_ops) {
  ev_timer* timer = (ev_timer*) malloc(sizeof (ev_timer));
  if (!timer) {
    LOG("ERROR: failed to allocate timer");
    return NULL;
  }
  memset(timer, 0, sizeof (ev_timer));
  return (void*) timer;
}

void delete_timer(ev2citrusleaf_ev_operations* ev_ops, void* object) {
  if (!ev_ops || !object)
    return;

  ev_plugin_cookie_t* cookie = (ev_plugin_cookie_t*) ev_ops->cookie;
  ev_timer* timer = (ev_timer*) object;

  if (ev_is_active(timer)) {
    ev_timer_stop(cookie->loop, timer);
  }
}

void destroy_timer(ev2citrusleaf_ev_operations* ev_ops, void* object) {
  if (!ev_ops || !object)
    return;

  ev_plugin_cookie_t* cookie = (ev_plugin_cookie_t*) ev_ops->cookie;
  ev_timer* timer = (ev_timer*) object;

  if (ev_is_active(timer)) {
    ev_timer_stop(cookie->loop, timer);
  }

  if (timer->data) {
    ev_callback_cookie_t* callback_cookie = (ev_callback_cookie_t*)
        timer->data;
    free(callback_cookie);
    timer->data = NULL;
  }

  free(timer);
}

int update_timer(ev2citrusleaf_ev_operations* ev_ops, void* object,
    int timeout_us, void* user_cookie, handler_t callback) {
  if (!ev_ops || !object)
    return -1;

  ev_plugin_cookie_t* cookie = (ev_plugin_cookie_t*) ev_ops->cookie;
  ev_timer* timer = (ev_timer*) object;
  if (ev_is_active(timer)) {
    ev_timer_stop(cookie->loop, timer);
  }

  double timeout = (double) timeout_us;
  timeout /= (double) 1000000.;
  ev_timer_init(timer, timer_callback, timeout, 0.);

  ev_callback_cookie_t* callback_cookie = (ev_callback_cookie_t*) timer->data;
  if (!callback_cookie) {
    callback_cookie = (ev_callback_cookie_t*)
        malloc(sizeof (ev_callback_cookie_t));
  }

  callback_cookie->cl_lib_callback = callback;
  callback_cookie->cl_lib_data = user_cookie;
  timer->data = (void*) callback_cookie;

  ev_timer_start(cookie->loop, timer);

  return 0;
}

void timer_callback(struct ev_loop* loop, ev_timer* timer, int events) {
  if (!timer)
    return;

  ev_callback_cookie_t* callback_cookie = (ev_callback_cookie_t*) timer->data;

  if (!callback_cookie || !callback_cookie->cl_lib_callback)
    return;

  callback_cookie->cl_lib_callback(-1, libev_events_to_cl(events),
      callback_cookie->cl_lib_data);
}

void* create_event(ev2citrusleaf_ev_operations* ev_ops) {
  ev_io* io = (ev_io*) malloc(sizeof (ev_io));
  if (!io) {
    LOG("ERROR: Failed to allocate IO event");
    return NULL;
  }
  memset(io, 0, sizeof (ev_io));
  return (void*) io;
}

void delete_event(ev2citrusleaf_ev_operations* ev_ops, int fd, void* object) {
  if (!ev_ops || !object)
    return;

  ev_plugin_cookie_t* cookie = (ev_plugin_cookie_t*) ev_ops->cookie;
  ev_io* io = (ev_io*) object;

  if (ev_is_active(io)) {
    ev_io_stop(cookie->loop, io);
  }
}

void destroy_event(ev2citrusleaf_ev_operations* ev_ops, void* object) {
  if (!ev_ops || !object)
    return;

  ev_plugin_cookie_t* cookie = (ev_plugin_cookie_t*) ev_ops->cookie;
  ev_io* io = (ev_io*) object;

  if (ev_is_active(io)) {
    ev_io_stop(cookie->loop, io);
  }

  if (io->data) {
    ev_callback_cookie_t* callback_cookie = (ev_callback_cookie_t*)
        io->data;
    free(callback_cookie);
    io->data = NULL;
  }

  free(io);
}

int update_event(ev2citrusleaf_ev_operations* ev_ops, int fd, void* object,
    short events, void* user_cookie, handler_t callback) {
  if (!ev_ops || !object)
    return -1;

  ev_plugin_cookie_t* cookie = (ev_plugin_cookie_t*) ev_ops->cookie;
  ev_io* io = (ev_io*) object;

  if (ev_is_active(io)) {
    ev_io_stop(cookie->loop, io);
  }

  ev_io_init(io, event_callback, fd, cl_events_to_libev(events));

  ev_callback_cookie_t* callback_cookie = (ev_callback_cookie_t*) io->data;
  if (!callback_cookie) {
    callback_cookie = (ev_callback_cookie_t*)
        malloc(sizeof (ev_callback_cookie_t));
  }

  callback_cookie->cl_lib_callback = callback;
  callback_cookie->cl_lib_data = user_cookie;
  io->data = (void*) callback_cookie;

  ev_io_start(cookie->loop, io);

  return 0;
}

void event_callback(struct ev_loop* loop, ev_io* io, int events) {
  if (!io)
    return;

  ev_callback_cookie_t* callback_cookie = (ev_callback_cookie_t*) io->data;

  if (!callback_cookie || !callback_cookie->cl_lib_callback)
    return;

  callback_cookie->cl_lib_callback(io->fd, libev_events_to_cl(events),
      callback_cookie->cl_lib_data);
}

short libev_events_to_cl(int events) {
  short cl_events = EV2CITRUSLEAF_EV_NONE;

  if (events & EV_ERROR) {
    return EV2CITRUSLEAF_EV_ERROR;;
  }

  if (events & EV_WRITE) {
    cl_events |= EV2CITRUSLEAF_EV_WRITE;
  }

  if (events & EV_READ) {
    cl_events |= EV2CITRUSLEAF_EV_READ;
  }

  if (events & EV_TIMER) {
    cl_events |= EV2CITRUSLEAF_EV_TIMER;
  }

  return cl_events;
}

int cl_events_to_libev(short events) {
  short ev_events = EV_NONE;

  if (events & EV2CITRUSLEAF_EV_ERROR) {
    return EV_ERROR;
  }

  if (events & EV2CITRUSLEAF_EV_WRITE) {
    ev_events |= EV_WRITE;
  }

  if (events & EV2CITRUSLEAF_EV_READ) {
    ev_events |= EV_READ;
  }

  if (events & EV2CITRUSLEAF_EV_TIMER) {
    ev_events |= EV_TIMER;
  }

  return ev_events;
}

void run_event_loop(ev2citrusleaf_ev_operations* ev_ops) {
  if (!ev_ops) {
    LOG("ERROR: NULL pointer passed to %s", __func__);
    return;
  }

  ev_plugin_cookie_t* cookie = (ev_plugin_cookie_t*) ev_ops->cookie;
  ev_run(cookie->loop, 0);
}

void stop_event_loop(ev2citrusleaf_ev_operations* ev_ops) {
  if (!ev_ops) {
    LOG("ERROR: NULL pointer passed to %s", __func__);
    return;
  }

  ev_plugin_cookie_t* cookie = (ev_plugin_cookie_t*) ev_ops->cookie;
  ev_break(cookie->loop, EVBREAK_ALL);
}

// ============================================================================
// Config and command line options implementation
// ============================================================================
static bool set_config(int argc, char** argv) {
  g_config.host = DEFAULT_HOST;
  g_config.port = DEFAULT_PORT;
  g_config.namespace = DEFAULT_NAMESPACE;
  g_config.set = DEFAULT_SET;
  g_config.timeout_msec = DEFAULT_TIMEOUT_MSEC;

  int c;
  while ((c = getopt(argc, argv, "h:p:n:s:m:")) != -1) {
    switch (c) {
    case 'h':
      g_config.host = strdup(optarg);
      break;

    case 'p':
      g_config.port = atoi(optarg);
      break;

    case 'n':
      g_config.namespace = strdup(optarg);
      break;

    case 's':
      g_config.set = strdup(optarg);
      break;

    case 'm':
      g_config.timeout_msec = atoi(optarg);
      break;

    default:
      destroy_config();
      usage();
      return false;
    }
  }

  LOG("host:                %s", g_config.host);
  LOG("port:                %d", g_config.port);
  LOG("namespace:           %s", g_config.namespace);
  LOG("set name:            %s", g_config.set);
  LOG("transaction timeout: %d msec", g_config.timeout_msec);

  return true;
}

static void destroy_config() {
  if (DEFAULT_HOST != g_config.host) {
    free((void*) g_config.host);
  }

  if (DEFAULT_NAMESPACE != g_config.namespace) {
    free((void*) g_config.namespace);
  }

  if (DEFAULT_SET != g_config.set) {
    free((void*) g_config.set);
  }
}

// ============================================================================
// CitrusLeaf
// ============================================================================
static bool start_cluster_management(ev2citrusleaf_ev_operations* ev_ops) {
  // Initialize CitrusLeaf client
  int result = ev2citrusleaf_init(NULL);

  if (0 != result) {
    LOG("ERROR: initializing client [%d]", result);
    return false;
  }

  // Create cluster handle
  g_cluster = ev2citrusleaf_cluster_create(ev_ops, NULL);

  if (!g_cluster) {
    LOG("ERROR: creating cluster handle");
    ev2citrusleaf_shutdown(true);
    return false;
  }

  // Connect to server
  result = ev2citrusleaf_cluster_add_host(g_cluster, (char*) g_config.host,
      g_config.port);

  if (0 != result) {
    LOG("ERROR: adding host to cluster [%d]", result);
    ev2citrusleaf_cluster_destroy(g_cluster);
    g_cluster = NULL;
    ev2citrusleaf_shutdown(true);
    return false;
  }

  void* timer = ev_ops->create_timer(ev_ops);

  if (!timer) {
    LOG("ERROR: allocating timer");
    ev2citrusleaf_cluster_destroy(g_cluster);
    g_cluster = NULL;
    ev2citrusleaf_shutdown(true);
    return false;
  }

  verify_cluster_callback_cookie_t* cookie = (verify_cluster_callback_cookie_t*)
      malloc(sizeof (verify_cluster_callback_cookie_t));
  cookie->event_ops = ev_ops;
  cookie->timer = timer;

  if (0 != ev_ops->update_timer(ev_ops, timer, CLUSTER_VERIFY_INTERVAL, cookie,
      verify_cluster_callback)) {
    LOG("ERROR: verifying cluster");
    ev_ops->destroy_timer(ev_ops, timer);
    ev2citrusleaf_cluster_destroy(g_cluster);
    g_cluster = NULL;
    ev2citrusleaf_shutdown(true);
    return false;
  }

  return true;
}

static void stop_cluster_management() {
  if (g_cluster) {
    ev2citrusleaf_cluster_destroy(g_cluster);
    g_cluster = NULL;
    ev2citrusleaf_shutdown(false);
  }
}

static void verify_cluster_callback(int fd, short events, void* udata) {
  static int n_prev = 0;
  static int tries = 0;
  int n;
  verify_cluster_callback_cookie_t* cookie =
      (verify_cluster_callback_cookie_t*) udata;

  cookie->event_ops->delete_timer(cookie->event_ops, cookie->timer);

  if (tries++ < CLUSTER_VERIFY_TRIES) {
    n = ev2citrusleaf_cluster_get_active_node_count(g_cluster);
    if (n > 0 && n == n_prev) {
      LOG("INFO: Found %d cluster node%s", n, n > 1 ? "s" : "");
      cookie->event_ops->destroy_timer(cookie->event_ops, cookie->timer);
      free(cookie);
      //XXX: Trigger operations
      return;
    }

    n_prev = n;
  } else {
    LOG("FATAL: cluster not connected");
    cookie->event_ops->destroy_timer(cookie->event_ops, cookie->timer);
    free(cookie);
    stop_cluster_management();
    return;
  }

  cookie->event_ops->update_timer(cookie->event_ops, cookie->timer,
      CLUSTER_VERIFY_INTERVAL, cookie, verify_cluster_callback);
}
